#!/bin/env python3


import xml.etree.ElementTree as ET
import sys
import os
import itertools
import re
import collections


class NCBITaxon:
    def __init__(self, taxid, rank, division_id):
        self.taxid = taxid
        self.rank = rank
        self.division_id = division_id

    def q(self, taxid=None, rank=None, division_id='0'):
        return (taxid is None or taxid == self.taxid) and (rank is None or rank == self.rank) and (division_id is None or division_id == self.division_id)


class NCBIName:
    def __init__(self, taxon, name_type):
        self.taxon = taxon
        self.name_type = name_type

    def q(self, taxid=None, rank=None, name_type=None):
        return self.taxon.q(taxid, rank) and (name_type is None or name_type == self.name_type)


def read_ncbi_nodes(filename):
    result = {}
    with open(filename) as f:
        for line in f:
            cols = line.split('\t|\t')
            taxid = 'ncbi:' + cols[0]
            rank = cols[2]
            division_id = cols[4]
            result[taxid] = NCBITaxon(taxid, rank, division_id)
    return result


def read_ncbi_names(nodes, filename):
    result = {}
    with open(filename) as f:
        for line in f:
            cols = line.split('\t|\t')
            taxid = 'ncbi:' + cols[0]
            name = cols[1]
            name_type = cols[3].replace('\t|\n', '')
            result[name] = NCBIName(nodes[taxid], name_type)
    return result


class BacDiveEntry:
    def __init__(self, ncbi_taxa, filename):
        self.filename = filename
        self.bacdive_id = os.path.basename(filename).replace('.xml', '')
        self.root = ET.parse(filename).getroot()
        self.strain_number = []
        self.species = None
        self.is_subspecies = False
        self.genus = None
        self.family = None
        self.ordo = None
        self.class_ = None
        self.phylum = None
        self.domain = None
        self.strain_number = tuple(self._get_strain_numbers())
        self._set_taxonomy()
        self.complete_name = tuple('%s %s' % (self.species, sn) for sn in self.strain_number)
        self.dispatch, self.ncbi_taxon = self._dispatch(ncbi_taxa)
        self.strain_taxid = self._get_strain_taxid()
        self.canonical = self._get_canonical()

    def _get_strain_numbers(self):
        for sne in self.root.iterfind('./strain_availability/strains/list-item/strain_number'):
            if sne.text is None:
                continue
            yield from (' '.join(snt.strip().split()) for snt in re.split('[,\n]', sne.text))
        desig = self.root.findtext('./taxonomy_name/strains/list-item/designation')
        if desig is not None and desig != '':
            for d in re.split('[,;]', desig):
                d = d.strip()
                if d.startswith('ST') and d.endswith('(HKI)'):
                    n = d.replace('ST', '', 1).replace('(HKI)', '')
                    yield 'ST ' + n
                    yield 'ST' + n
                yield d

    def _set_taxonomy(self):
        tax = self.root.find('./taxonomy_name/strains/list-item')
        if tax is None:
            return
        self.species = tax.findtext('./species')
        self.is_subspecies = bool(tax.findtext('./subspecies_epithet'))
        self.genus = tax.findtext('./genus')
        self.family = tax.findtext('./family')
        self.ordo = tax.findtext('./ordo')
        self.class_ = tax.findtext('./class')
        self.phylum = tax.findtext('./phylum')
        self.domain = tax.findtext('./domain')

    def _get_strain_taxid(self):
        return 'bd:' + self.bacdive_id

    def _get_canonical(self):
        for sn in self.strain_number:
            if sn.startswith('DSM'):
                return '%s %s' % (self.species, sn)
        if len(self.strain_number) > 0:
            return '%s %s' % (self.species, sn)
        return None

    def _match(self, ncbi_taxa, s, rank=None, name_type=None):
        if s in ncbi_taxa:
            name = ncbi_taxa[s]
            if name.q(rank=rank, name_type=name_type):
                return name
        return None

    def _match_first(self, ncbi_taxa, rank=None, name_type=None):
        for n in itertools.chain(self.strain_number, self.complete_name):
            name = self._match(ncbi_taxa, n, rank, name_type)
            if name is not None:
                return name

    def _dispatch(self, ncbi_taxa):
        if len(self.strain_number) == 0:
            return 'no-number', None

        name = self._match_first(ncbi_taxa, 'no rank')
        if name is not None:
            return 'equivalent', name.taxon

        name = self._match_first(ncbi_taxa, name_type='type material')
        if name is not None:
            return 'type-material-of-' + name.taxon.rank, name.taxon

        rank = 'subspecies' if self.is_subspecies else 'species'
        name = self._match(ncbi_taxa, self.species, rank)
        if name is not None:
            return 'append-to-' + rank, name.taxon

        if rank == 'subspecies':
            name = self._match(ncbi_taxa, self.genus, 'species')
            if name is not None:
                return 'append-to-species', name.taxon
        name = self._match(ncbi_taxa, self.genus, 'genus')
        if name is not None:
            return 'append-to-genus', name.taxon
        name = self._match(ncbi_taxa, self.family, 'family')
        if name is not None:
            return 'append-to-family', name.taxon
        name = self._match(ncbi_taxa, self.ordo, 'order')
        if name is not None:
            return 'append-to-order', name.taxon
        name = self._match(ncbi_taxa, self.class_, 'class')
        if name is not None:
            return 'append-to-class', name.taxon
        name = self._match(ncbi_taxa, self.phylum, 'phylum')
        if name is not None:
            return 'append-phylum', name.taxon
        name = self._match(ncbi_taxa, self.domain, 'superkingdom')
        if name is not None:
            return 'append-to-superkingdom', name.taxon
        return 'fail', None


_, NCBI_NODES_FILE, NCBI_NAMES_FILE, INPUT_DIR, OUTPUT_DIR = sys.argv


def of(fn):
    return open(os.path.join(OUTPUT_DIR, fn), 'w')


def line(f, sep, *cols, tail=None):
    f.write(sep.join(str(i) for i in cols))
    if tail is not None:
        f.write(tail)
    f.write('\n')


sys.stderr.write('Reading NCBI taxonomy: %s %s\n' % (NCBI_NODES_FILE, NCBI_NAMES_FILE))
NCBI_TAXA = read_ncbi_names(read_ncbi_nodes(NCBI_NODES_FILE), NCBI_NAMES_FILE)
os.makedirs(OUTPUT_DIR, exist_ok=True)
with of('dispatch-report.txt') as out_dispatch, of('dsmz-nodes.dmp') as out_nodes, of('dsmz-names.dmp') as out_names, of('bacdive-to-taxid.txt') as out_link, of('warnings.txt') as out_warn:
    sys.stderr.write('Reading BacDive entries: %s\n' % INPUT_DIR)
    count_entries = 0
    count_dispatches = collections.defaultdict(int)
    count_new_nodes = 0
    count_names_of_new_nodes = 0
    count_new_synonyms = 0
    new_ids = set()
    for dirpath, _, filenames in os.walk(INPUT_DIR):
        for fn in filenames:
            if not fn.endswith('.xml'):
                continue
            e = BacDiveEntry(NCBI_TAXA, os.path.join(dirpath, fn))
            count_entries += 1
            count_dispatches[e.dispatch] += 1
            line(out_dispatch, '\t', e.filename, e.dispatch, '' if e.ncbi_taxon is None else e.ncbi_taxon.taxid)
            if e.dispatch.startswith('append-to-'):
                if e.strain_taxid in new_ids:
                    line(out_warn, '\t', e.filename, 'duplicate id ' + e.strain_taxid)
                    continue
                if e.ncbi_taxon is not None and e.ncbi_taxon.division_id != '0':
                    line(out_warn, '\t', e.filename, 'not bacteria (div %s) (strain %s) (taxid %s)' % (e.ncbi_taxon.division_id, e.strain_taxid, e.ncbi_taxon.taxid))
                new_ids.add(e.strain_taxid)
                line(out_nodes, '\t|\t',
                     e.strain_taxid,  # taxid
                     e.ncbi_taxon.taxid,  # parent taxid
                     'no rank',  # rank
                     '',  # embl code
                     e.ncbi_taxon.division_id,   # division id
                     1,   # inherited div flag
                     11,  # genetic code id
                     1,   # inherited GC  flag
                     0,   # mitochondrial genetic code id
                     1,   # inherited MGC flag
                     1,   # GenBank hidden flag
                     0,   # hidden subtree root flag
                     '',  # comments
                     tail='\t|')
                line(out_link, '\t', e.bacdive_id, e.strain_taxid)
                count_new_nodes += 1
                if e.canonical is None:
                    line(out_warn, '\t', e.filename, 'no canonical')
                else:
                    for name in itertools.chain(e.strain_number, e.complete_name):
                        line(out_names, '\t|\t', e.strain_taxid, name, '', 'scientific name' if name == e.canonical else 'strain number', tail='\t|')
                        count_names_of_new_nodes += 1
            elif e.dispatch == 'equivalent' or e.dispatch.startswith('type-material-of-'):
                line(out_link, '\t', e.bacdive_id, e.ncbi_taxon.taxid)
                for name in itertools.chain(e.strain_number, e.complete_name):
                    if name not in NCBI_TAXA:
                        line(out_names, '\t|\t', e.ncbi_taxon.taxid, name, '', 'strain number', tail='\t|')
                        count_new_synonyms += 1
    sys.stderr.write('Entries: %d\n' % count_entries)
    sys.stderr.write('Dispatches:\n')
    for c in count_dispatches.items():
        sys.stderr.write('  %s: %d\n' % c)
    sys.stderr.write('New nodes: %d\n' % count_new_nodes)
    sys.stderr.write('Names of new nodes: %d\n' % count_names_of_new_nodes)
    sys.stderr.write('New synonyms: %d\n' % count_new_synonyms)
